@@ -1,8 +1,8 @@
 #!/usr/bin/python
 #
-# Author: Jashua R. Cloutier (jashuac@bellsouth.com)
+# Author: Jashua R. Cloutier (contact via sourceforge username:senexcanis)
 #
-# Copyright (C) 2009, Jashua R. Cloutier
+# Copyright (C) 2010, Jashua R. Cloutier
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -48,9 +48,17 @@ representing the class
 import ply.lex as lex
 import os
 import sys
+import re
 
-__version__ = "1.05"
+import inspect
 
+def lineno():
+    """Returns the current line number in our program."""
+    return inspect.currentframe().f_back.f_lineno
+
+
+__version__ = "1.8"
+version = "1.8"
 tokens = [
     'NUMBER',
     'NAME',
@@ -77,9 +85,9 @@ tokens = [
     'NEW_LINE',
 ]
 
-t_ignore = " \t\r~[].|!?%@"
+t_ignore = " \t\r[].|!?%@"
 t_NUMBER = r'[0-9][0-9XxA-Fa-f]*'
-t_NAME = r'[<>A-Za-z_][A-Za-z0-9_]*'
+t_NAME = r'[<>A-Za-z_~][A-Za-z0-9_]*'
 t_OPERATOR_DIVIDE_OVERLOAD = r'/='
 t_OPEN_PAREN = r'\('
 t_CLOSE_PAREN = r'\)'
@@ -90,7 +98,16 @@ t_COLON = r':'
 t_COMMA = r','
 t_PRECOMP_MACRO = r'\#.*'
 t_PRECOMP_MACRO_CONT = r'.*\\\n'
-t_COMMENT_SINGLELINE = r'\/\/.*\n'
+def t_COMMENT_SINGLELINE(t):
+    r'\/\/.*\n'
+    global doxygenCommentCache
+    if t.value.startswith("///") or t.value.startswith("//!"):
+        if doxygenCommentCache:
+            doxygenCommentCache += "\n"
+        if t.value.endswith("\n"):
+            doxygenCommentCache += t.value[:-1]
+        else:
+            doxygenCommentCache += t.value
 t_ASTERISK = r'\*'
 t_MINUS = r'\-'
 t_PLUS = r'\+'
@@ -102,7 +119,15 @@ t_CHAR_LITERAL = "'.'"
 #TODO: This does not work with the string "bla \" bla"
 t_STRING_LITERAL = r'"([^"\\]|\\.)*"'
 #Found at http://ostermiller.org/findcomment.html
-t_COMMENT_MULTILINE = r'/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/'
+def t_COMMENT_MULTILINE(t):
+    r'/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/'
+    global doxygenCommentCache
+    if t.value.startswith("/**") or t.value.startswith("/*!"):
+        #not sure why, but get double new lines
+        v = t.value.replace("\n\n", "\n")
+        #strip prefixing whitespace
+        v = re.sub("\n[\s]+\*", "\n*", v)
+        doxygenCommentCache += v
 def t_NEWLINE(t):
     r'\n+'
     t.lexer.lineno += len(t.value)
@@ -119,6 +144,16 @@ supportedAccessSpecifier = [
     'private'
 ]
 
+doxygenCommentCache = ""
+
+def is_namespace(nameStack):
+    """Determines if a namespace is being specified"""
+    if len(nameStack) == 0:
+        return False
+    if nameStack[0] == "namespace":
+        return True
+    return False
+
 def is_enum_namestack(nameStack):
     """Determines if a namestack is an enum namestack"""
     if len(nameStack) == 0:
@@ -136,6 +171,7 @@ class CppClass(dict):
     
     Contains the following Keys:
     self['name'] - Name of the class
+    self['doxygen'] - Doxygen comments associated with the class if they exist
     self['inherits'] - List of Classes that this one inherits where the values
         are of the form {"access": Anything in supportedAccessSpecifier
                                   "class": Name of the class
@@ -172,9 +208,14 @@ class CppClass(dict):
     }
     """
     def __init__(self, nameStack):
+        if (debug): print "Class:   ",  nameStack
         if (len(nameStack) < 2):
             print "Error detecting class"
             return
+        global doxygenCommentCache
+        if len(doxygenCommentCache):
+            self["doxygen"] = doxygenCommentCache
+            doxygenCommentCache = ""
         self["name"] = nameStack[1]
         inheritList = []
         if ":" in nameStack:
@@ -211,10 +252,16 @@ class CppClass(dict):
         self['methods'] = methodAccessSpecificList
         self['properties'] = propertyAccessSpecificList
         self['enums'] = enumAccessSpecificList
+        self['namespace'] = ""
 
     def __repr__(self):
         """Convert class to a string"""
-        rtn = "class %s\n"%(self["name"])
+        namespace_prefix = ""
+        if self["namespace"]: namespace_prefix = self["namespace"] + "::"
+        rtn = "class %s\n"%(namespace_prefix + self["name"])
+        try:
+            print self["doxygen"], 
+        except: pass
         if "inherits" in self.keys():
             rtn += "Inherits: "
             for inheritClass in self["inherits"]:
@@ -247,26 +294,64 @@ class CppMethod(dict):
     Contains the following Keys:
     self['rtnType'] - Return type of the method (ex. "int")
     self['name'] - Name of the method (ex. "getSize")
+    self['doxygen'] - Doxygen comments associated with the method if they exist
     self['parameters'] - List of CppVariables
     """
-    def __init__(self, nameStack):
+    def __init__(self, nameStack, curClass):
         if (debug): print "Method:   ",  nameStack
-        self["rtnType"] = " ".join(nameStack[:nameStack.index('(') - 1])
-        if len(self["rtnType"]) == 0:
+        global doxygenCommentCache
+        if len(doxygenCommentCache):
+            self["doxygen"] = doxygenCommentCache
+            doxygenCommentCache = ""
+        if "operator" in nameStack:
+            self["rtnType"] = " ".join(nameStack[:nameStack.index('operator')])
+            self["name"] = "".join(nameStack[nameStack.index('operator'):nameStack.index('(')])
+        else:
+            self["rtnType"] = " ".join(nameStack[:nameStack.index('(') - 1])
+            self["name"] = " ".join(nameStack[nameStack.index('(') - 1:nameStack.index('(')])
+        if len(self["rtnType"]) == 0 or self["name"] == curClass:
             self["rtnType"] = "void"
-        self["name"] = " ".join(nameStack[nameStack.index('(') - 1:nameStack.index('(')])
         paramsStack = nameStack[nameStack.index('(') + 1: ]
         #Remove things from the stack till we hit the last paren, this helps handle abstract and normal methods
         while (paramsStack[-1]  != ")"):
             paramsStack.pop()
         paramsStack.pop()
         params = []
+        #See if there is a doxygen comment for the variable
+        doxyVarDesc = {}
+        #TODO: Put this into a class
+        if self.has_key("doxygen"):
+            doxyLines = self["doxygen"].split("\n")
+            lastParamDesc = ""
+            for doxyLine in doxyLines:
+                if " @param " in doxyLine or " \param " in doxyLine:
+                    try:
+                        #Strip out the param
+                        doxyLine = doxyLine[doxyLine.find("param ") + 6:]
+                        (var, desc) = doxyLine.split(" ", 1)
+                        doxyVarDesc[var] = desc.strip()
+                        lastParamDesc = var
+                    except: pass
+                elif " @return " in doxyLine or " \return " in doxyLine:
+                    lastParamDesc = ""
+                    # not handled for now
+                elif lastParamDesc:
+                    try:
+                        doxyLine = doxyLine.strip()
+                        if " " not in doxyLine:
+                            lastParamDesc = ""
+                            continue
+                        doxyLine = doxyLine[doxyLine.find(" ") + 1:]
+                        doxyVarDesc[lastParamDesc] += " " + doxyLine
+                    except: pass
+        
+        #Create the variable now
         while (len(paramsStack)):
             if (',' in paramsStack):
-                params.append(CppVariable(paramsStack[0:paramsStack.index(',')]))
+                params.append(CppVariable(paramsStack[0:paramsStack.index(',')],  doxyVarDesc=doxyVarDesc))
                 paramsStack = paramsStack[paramsStack.index(',') + 1:]
             else:
-                param = CppVariable(paramsStack)
+                param = CppVariable(paramsStack,  doxyVarDesc=doxyVarDesc)
                 if len(param.keys()):
                     params.append(param)
                 break
@@ -279,13 +364,20 @@ class CppVariable(dict):
     Contains the following Keys:
     self['type'] - Type for the variable (ex. "const string &")
     self['name'] - Name of the variable (ex. "numItems")
+    self['namespace'] - Namespace containing the enum
+    self['desc'] - Description of the variable if part of a method (optional)
+    self['doxygen'] - Doxygen comments associated with the method if they exist
     self['defaltValue'] - Default value of the variable, this key will only
         exist if there is a default value
     """
-    def __init__(self, nameStack):
+    def __init__(self, nameStack,  **kwargs):
         if (debug): print "Variable: ",  nameStack
         if (len(nameStack) < 2):
             return
+        global doxygenCommentCache
+        if len(doxygenCommentCache):
+            self["doxygen"] = doxygenCommentCache
+            doxygenCommentCache = ""
         if ("=" in nameStack):
             self["type"] = " ".join(nameStack[:nameStack.index("=") - 1])
             self["name"] = nameStack[nameStack.index("=") - 1]
@@ -297,12 +389,17 @@ class CppVariable(dict):
         self["type"] = self["type"].replace(": ",":")
         self["type"] = self["type"].replace(" <","<")
         self["type"] = self["type"].replace(" >",">")
+        #Optional doxygen description
+        try:
+            self["desc"] = kwargs["doxyVarDesc"][self["name"]]
+        except: pass
 
 class CppEnum(dict):
     """Takes a name stack and turns it into an Enum
     
     Contains the following Keys:
     self['name'] - Name of the enum (ex. "ItemState")
+    self['namespace'] - Namespace containing the enum
     self['values'] - List of values where the values are a dictionary of the
         form {"name": name of the key (ex. "PARSING_HEADER"),
                   "value": Specified value of the enum, this key will only exist
@@ -313,6 +410,10 @@ class CppEnum(dict):
         if len(nameStack) < 4 or "{" not in nameStack or "}" not in nameStack:
             #Not enough stuff for an enum
             return
+        global doxygenCommentCache
+        if len(doxygenCommentCache):
+            self["doxygen"] = doxygenCommentCache
+            doxygenCommentCache = ""
         valueList = []
         #Figure out what values it has
         valueStack = nameStack[nameStack.index('{') + 1: nameStack.index('}')]
@@ -328,6 +429,9 @@ class CppEnum(dict):
                 valueList.append({"name": tmpStack[0]})
             elif len(tmpStack) >= 3 and tmpStack[1] == "=":
                 valueList.append({"name": tmpStack[0], "value": " ".join(tmpStack[2:])})
+            elif len(tmpStack) == 2 and tmpStack[1] == "=":
+                if (debug): print "Missed value for %s"%tmpStack[0]
+                valueList.append({"name": tmpStack[0]})
         if len(valueList):
             self["values"] = valueList
         else:
@@ -346,7 +450,8 @@ class CppEnum(dict):
             for var in postBraceStack:
                 if "," in var:
                     continue
-                self["instances"].append(var)                
+                self["instances"].append(var)
+        self["namespace"] = ""
 
 class CppHeader:
     """Parsed C++ class header
@@ -357,11 +462,11 @@ class CppHeader:
     """
     def __init__(self, headerFileName, argType = "file"):
         if (argType == "file"):
-            self.headerFileName = headerFileName
+            self.headerFileName = os.path.expandvars(headerFileName)
             self.mainClass = os.path.split(self.headerFileName)[1][:-2]
             headerFileStr = ""
-            if headerFileName[-2:] != ".h":
-                raise Exception("file must be a header file and end with .h")
+#            if headerFileName[-2:] != ".h":
+#                raise Exception("file must be a header file and end with .h")
         elif argType == "string":
             self.headerFileName = ""
             self.mainClass = "???"
@@ -370,7 +475,9 @@ class CppHeader:
             raise Exception("Arg type must be either file or string")
         self.curClass = ""
         self.classes = {}
+        self.enums = []
         self.nameStack = []
+        self.nameSpaces = []
         self.curAccessSpecifier = 'private'
     
         if (len(self.headerFileName)):
@@ -389,16 +496,22 @@ class CppHeader:
                 curLine = tok.lineno
                 curChar = tok.lexpos
                 if (tok.type == 'OPEN_BRACE'):
+                    if len(self.nameStack) and is_namespace(self.nameStack):
+                        self.nameSpaces.append(self.nameStack[1])
                     if len(self.nameStack) and not is_enum_namestack(self.nameStack):
-                        self.evaluateStack()
+                        self.evaluate_stack()
                     else:
                         self.nameStack.append(tok.value)
                     self.braceDepth += 1
                 elif (tok.type == 'CLOSE_BRACE'):
+                    if self.braceDepth == 0:
+                        continue
+                    if (self.braceDepth == len(self.nameSpaces)):
+                        tmp = self.nameSpaces.pop()
                     if len(self.nameStack) and is_enum_namestack(self.nameStack):
                         self.nameStack.append(tok.value)
-                    elif self.braceDepth < 2:
-                        self.evaluateStack()
+                    elif self.braceDepth < 10:
+                        self.evaluate_stack()
                     else:
                         self.nameStack = []
                     self.braceDepth -= 1
@@ -424,7 +537,7 @@ class CppHeader:
                 elif (tok.type == 'NAME' or tok.type == 'AMPERSTAND' or tok.type == 'ASTERISK'):
                     if (tok.value == 'class'):
                         self.nameStack.append(tok.value)
-                    elif (tok.value in supportedAccessSpecifier and self.braceDepth == 1):
+                    elif (tok.value in supportedAccessSpecifier and self.braceDepth == len(self.nameSpaces) + 1):
                         self.curAccessSpecifier = tok.value
                     else:
                         self.nameStack.append(tok.value)
@@ -434,45 +547,68 @@ class CppHeader:
                         continue
                     self.nameStack.append(tok.value)
                 elif (tok.type == 'SEMI_COLON'):
-                    if (self.braceDepth < 2):
-                        self.evaluateStack()
+                    if (self.braceDepth < 10):
+                        self.evaluate_stack()
         except:
             raise CppParseError("Not able to parse %s on line %d evaluating \"%s\"\nError around: %s"
                                 % (self.headerFileName, tok.lineno, tok.value, " ".join(self.nameStack)))
         
-    def evaluateStack(self):
+    def evaluate_stack(self):
         """Evaluates the current name stack"""
+        global doxygenCommentCache
+        if (debug): print "Evaluating stack %s at..."%self.nameStack
         if (len(self.curClass)):
             if (debug): print "%s (%s) "%(self.curClass, self.curAccessSpecifier), 
         if (len(self.nameStack) == 0):
+            if (debug): print "line ",lineno()
             if (debug): print "(Empty Stack)"
             return
+        elif (self.nameStack[0] == "namespace"):
+            #Taken care of outside of here
+            pass
         elif (self.nameStack[0] == "class"):
-            self.evaluateClassStack()
+            if (debug): print "line ",lineno()
+            self.evaluate_class_stack()
+        elif (self.nameStack[0] == "struct"):
+            if (debug): print "line ",lineno()
+            self.curAccessSpecifier = "public"
+            self.evaluate_class_stack()
         elif (len(self.curClass) == 0):
+            if (debug): print "line ",lineno()
+            if is_enum_namestack(self.nameStack):
+                self.evaluate_enum_stack()
             self.nameStack = []
+            doxygenCommentCache = ""
             return
         elif (self.braceDepth < 1):
+            if (debug): print "line ",lineno()
             #Ignore global stuff for now
             if (debug): print "Global stuff: ",  self.nameStack
             self.nameStack = []
+            doxygenCommentCache = ""
             return
-        elif (self.braceDepth > 1):
+        elif (self.braceDepth > len(self.nameSpaces) + 1):
+            if (debug): print "line ",lineno()
             self.nameStack = []
+            doxygenCommentCache = ""
             return
         elif is_enum_namestack(self.nameStack):
+            if (debug): print "line ",lineno()
             #elif self.nameStack[0] == "enum":
-            self.evaluateEnumStack()
+            self.evaluate_enum_stack()
         elif ('(' in self.nameStack):
-            self.evaluateMethodStack()
+            if (debug): print "line ",lineno()
+            self.evaluate_method_stack()
         else:
-            self.evaluatePropertyStack()
+            if (debug): print "line ",lineno()
+            self.evaluate_property_stack()
         self.nameStack = []
+        doxygenCommentCache = ""
     
-    def evaluateClassStack(self):
+    def evaluate_class_stack(self):
         """Create a Class out of the name stack (but not its parts)"""
         #dont support sub classes today
-        if self.braceDepth != 0:
+        if self.braceDepth != len(self.nameSpaces):
             return
         newClass = CppClass(self.nameStack)
         if len(newClass.keys()):
@@ -480,24 +616,31 @@ class CppHeader:
             self.classes[self.curClass] = newClass
         else:
             self.curClass = ""
+        newClass["namespace"] = self.cur_namespace()
 
-    def evaluateMethodStack(self):
+    def evaluate_method_stack(self):
         """Create a method out of the name stack"""
-        newMethod = CppMethod(self.nameStack)
+        newMethod = CppMethod(self.nameStack, self.curClass)
         if len(newMethod.keys()):
             self.classes[self.curClass]["methods"][self.curAccessSpecifier].append(newMethod)
     
-    def evaluatePropertyStack(self):
+    def evaluate_property_stack(self):
         """Create a Property out of the name stack"""
         newVar = CppVariable(self.nameStack)
         if len(newVar.keys()):
             self.classes[self.curClass]["properties"][self.curAccessSpecifier].append(newVar)
 
-    def evaluateEnumStack(self):
+    def evaluate_enum_stack(self):
         """Create an Enum out of the name stack"""
         newEnum = CppEnum(self.nameStack)
         if len(newEnum.keys()):
-            self.classes[self.curClass]["enums"][self.curAccessSpecifier].append(newEnum)
+            if len(self.curClass):
+                newEnum["namespace"] = self.cur_namespace()
+                self.classes[self.curClass]["enums"][self.curAccessSpecifier].append(newEnum)
+            else:
+                newEnum["namespace"] = self.cur_namespace()                            
+#                print "Adding global enum"
+                self.enums.append(newEnum)
             #This enum has instances, turn them into properties
             if newEnum.has_key("instances"):
                 instanceType = "enum"
@@ -505,9 +648,20 @@ class CppHeader:
                     instanceType = newEnum["name"]
                 for instance in newEnum["instances"]:
                     self.nameStack = [instanceType,  instance]
-                    self.evaluatePropertyStack()
+                    self.evaluate_property_stack()
                 del newEnum["instances"]
 
+    def cur_namespace(self, add_double_colon = False):
+        rtn = ""
+        i = 0
+        while i < len(self.nameSpaces):
+            rtn += self.nameSpaces[i]
+            if add_double_colon or i < len(self.nameSpaces) - 1:
+                rtn += "::"
+            i+=1
+        return rtn
+    
+    
     def __repr__(self):
         rtn = ""
         for className in self.classes.keys():
